import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)

/**
 * Write a description of class PathFindingWorld here.
 * 
 * @author (your name) 
 * @version (a version number or a date)
 */
public class PathFindingWorld extends World
{
    private byte[][][][] pathsToFrom;
    private int[][][][] distToFrom;
    private boolean[][] occupied; //Quick box lookup

    public static final byte STAY = -1;
    public static final byte LEFT = 1;
    public static final byte RIGHT = 2;
    public static final byte UP = 3;
    public static final byte DOWN = 4;
    
    private int toX, toY; // temporary variables

    /**
     * Constructor for objects of class PathFindingWorld.
     * 
     */
    public PathFindingWorld(int width, int height, int cellSize)
    {    
        super(width, height, cellSize);
        
        occupied = new boolean[getWidth()][getHeight()];
        pathsToFrom = new byte[getWidth()][getHeight()][getWidth()][getHeight()];
        distToFrom = new int[getWidth()][getHeight()][getWidth()][getHeight()];
        
    }
    
    public void setOccupied(int x, int y, boolean occ)
    {
        occupied[x][y] = occ;
    }
    
    public void calculatePaths()
    {
        for (toX = 0; toX < getWidth(); toX++)
        {
            for (toY = 0; toY < getHeight(); toY++)
            {
                calculatePathFrom(pathsToFrom[toX][toY], distToFrom[toX][toY], toX, toY, STAY, 0);
            }
        }
    }
    
    protected boolean outOfBounds(int x, int y)
    {
        return !(x >=0 && y >= 0 && x < getWidth() && y < getHeight());
    }
    
    /**
     * Each destination is processed with an independent call to this function.
     * 
     * The paths[][] is a grid of directions in which to head to reach the destination.
     * The distances[][] is a grid of shortest manhattan distances to reach the destination.
     * The fromX, fromY is the source position being considered for this destination.
     * The newDir and newDist are the latest options being considered for the source fromX, fromY.
     * 
     */
    private void calculatePathFrom(byte[][] paths, int[][] dist, int fromX, int fromY, byte newDir, int newDist)
    {
        if (outOfBounds(fromX, fromY) || occupied[fromX][fromY])
            return; //Outside world or blocked
            
        byte prevDir = paths[fromX][fromY];

        if (prevDir == STAY)
            return;
            
        int prevDist = dist[fromX][fromY];
        
        if (prevDir != 0 && prevDist != 0 && prevDist < newDist)
            return; //Already processed and already have shorter path
        
        if (prevDir != 0 && prevDist > 1 && prevDist == newDist
            && (prevDir == newDir || distanceSq(toX - moveX(fromX, newDir), toY - moveY(fromY, newDir)) > distanceSq(toX - moveX(fromX, prevDir), toY - moveY(fromY, prevDir))))
            return; //Already processed and already have shorter or equal path
        
        paths[fromX][fromY] = newDir;
        dist[fromX][fromY] = newDist;
        
        calculatePathFrom(paths, dist, fromX - 1, fromY, RIGHT, newDist + 1);
        calculatePathFrom(paths, dist, fromX + 1, fromY, LEFT, newDist + 1);
        calculatePathFrom(paths, dist, fromX, fromY - 1, UP, newDist + 1);
        calculatePathFrom(paths, dist, fromX, fromY + 1, DOWN, newDist + 1);
    }
    
    private static int distanceSq(int xdist, int ydist)
    {
        return xdist * xdist + ydist * ydist;
    }
    
    public static int moveX(int x, byte path)
    {
        switch (path)
        {
            case LEFT: return x - 1;
            case RIGHT: return x + 1;
            default: return x;
        }
    }
    
    public static int moveY(int y, byte path)
    {
        switch (path)
        {
            case UP: return y + 1;
            case DOWN: return y - 1;
            default: return y;
        }
    }
    
    public byte getDirToFrom(int toX, int toY, int fromX, int fromY)
    {
        return pathsToFrom[toX][toY][fromX][fromY];
    }
}
